import time
import weakref
import pygame
from spw import canvas


TICK = pygame.USEREVENT + 1


class Context(object):
    current = None
    def __init__(self, screen):
        self.screen = screen
        self.canvas = canvas.Canvas(self.screen)
        self.dirty = []
        self.cleanup_fn = None
        Context.current = self

    def __enter__(self):
        self.canvas.clear()
        if self.cleanup_fn:
            self.cleanup_fn(self.dirty)
        else:
            for r in self.dirty:
                self.screen.fill((0,0,0), r)
        return self.canvas

    def __exit__(self, type, value, traceback):
        dirty = self.canvas._rects.keys()
        pygame.display.update(self.dirty + dirty)
        self.dirty = dirty


class Expando(object):
    def get(self, key, default=None):
        try:        
            return self.__dict__[key]
        except KeyError:
            if callable(default):
                v = self.__dict__[key] = default()
            else:
                v = self.__dict__[key] = default
            return v


class Skin(object):
    font_file = None
    font_size = 16
    font_color = (128,255,128)

    def __init__(self):
        if self.font_file is None: self.font_file = pygame.font.get_default_font()
        self.font = pygame.font.Font(self.font_file, self.font_size)

    def draw_button(self, state, canvas, rect, label, hover_hot_active):
        hover, hot, active = hover_hot_active
        image = state.get("text:%s"%label, lambda:self.font.render(label, True, self.font_color))
        if hover:
            canvas.line_width(2)
        else:
            canvas.line_width(1)
        dr = canvas.quad(rect)
        r = image.get_rect()
        r.center = dr.center 
        canvas.blit(image, r)

    def draw_toggle(self, state, canvas, rect, hover_hot_mode):
        hover, hot, mode = hover_hot_mode
        if mode:
            canvas.line_width(0)
        elif hover:
            canvas.line_width(2)
        else:
            canvas.line_width(1)
        canvas.quad(rect)

    def draw_menu(self, state, canvas, rect, label, hover_hot):
        hover, hot = hover_hot
        image = state.get("text:%s"%label, lambda:self.font.render(label, True, self.font_color))
        if hover or hot:
            canvas.line_width(2)
        else:
            canvas.line_width(1)
        dr = canvas.quad(rect)
        r = image.get_rect()
        r.center = dr.center 
        canvas.blit(image, r)

    def draw_menu_item(self, state, canvas, rect, index_label):
        rect = pygame.Rect(rect)
        index, label = index_label
        image = state.get("text:%s"%label, lambda:self.font.render(label, True, self.font_color))
        canvas.line_width(1)
        canvas.quad(rect)
        r = image.get_rect()
        r.center = rect.center 
        canvas.blit(image, r)

    def draw_input(self, state, canvas, rect, hover_hot_active, text, cursor, pick):
        rect = pygame.Rect(rect)
        hover, hot, active = hover_hot_active
        image = self.font.render(text, True, self.font_color) 
        if active:
            canvas.color((28,255,28))
        canvas.quad(rect)
        canvas.clip(rect)
        cursor_offset = self.font.size(text[:cursor])[0]
        w,h = image.get_size() 
        if w > rect[2]:
            x = rect[0] - (w - rect[2])
        else:
            x = rect[0]
        y = rect[1] + ((rect[3] - h) / 2)

        if (x+cursor_offset) < rect[0]:
            one_char = self.font.size(text[cursor-1:cursor])[0]
            x = rect[0] - (cursor_offset - one_char)

        if pick is not None:
            A = (rect[0] - x) + (pick[0] - rect[0])
            for i in xrange(0, len(text)+1):
                w = self.font.size(text[:i])[0]
                if w > A:
                    state.cursor = i 
                    break
        if active:
            canvas.circle((x + cursor_offset, rect[1]+rect[3]), 5)
         
        canvas.blit(image, (x, y))


class GUI(object):
    frame = None
    hover = None
    hot = None
    active = None
    skin = None
    _state = weakref.WeakKeyDictionary()

    @classmethod
    def init(self):
        if self.skin is None: self.skin = Skin()
        pygame.time.set_timer(TICK, 100)

    @classmethod
    def draw(self, instance, canvas, event):
        self.frame = instance
        self.canvas = canvas
        self.event = event
        instance.draw_gui()
        self.frame = None
        self.canvas = None
        self.event = None

    @classmethod
    def STATE(self, name):
        states = self._state.setdefault(self.frame, {})
        return states.setdefault(name, Expando())

    @classmethod
    def ID(self, name):
        return id(self.frame), name

    @classmethod
    def clicked(self, id, rect):
        event = self.event
        canvas = self.canvas
        result = False
        if event.type == pygame.MOUSEMOTION:
            # if mouse is over me, set me to hover
            # if mouse is not over em, and i am set to hover, unset my hover
            hit = canvas.rect(rect).collidepoint(event.pos)
            if hit:
                GUI.hover = id
            elif GUI.hover == id:
                GUI.hover = None
        elif event.type == pygame.MOUSEBUTTONDOWN:
            # if i am hit, set me to hot
            hit = self.canvas.rect(rect).collidepoint(event.pos)
            if hit: GUI.hot = id
        elif event.type == pygame.MOUSEBUTTONUP:
            # if I am hot and I am hit set me to active
            # if I am hot set me to not hot 
            # if i am active and not hit, make me inactive
            hit = self.canvas.rect(rect).collidepoint(event.pos)
            if GUI.hot == id:
                if hit: GUI.active = id
                GUI.hot = None
            if GUI.active == id:
                if not hit: GUI.active = None
                
        return GUI.hover == id, GUI.hot == id, GUI.active == id

    @classmethod
    def button(self, name, label, rect):
        id = self.ID(name)
        state = self.STATE(name)
        hover, hot, active = self.clicked(id, rect)
        canvas = self.canvas
        canvas.push()
        self.skin.draw_button(state, canvas, rect, label, (hover, hot, active))
        canvas.pop()
        return active

    @classmethod
    def toggle(self, name, rect):
        id = self.ID(name)
        state = self.STATE(name)
        hover, hot, active = self.clicked(id, rect)
        mode = state.get("mode", False)
        if active: 
            mode = not mode
            GUI.active = None
        state.mode = mode
        canvas = self.canvas
        canvas.push()
        self.skin.draw_toggle(state, canvas, rect, (hover,hot,mode))
        canvas.pop()
        return mode

    @classmethod
    def menu(self, name, label, rect, options):
        id = self.ID(name)
        state = self.STATE(name)
        hover, hot, active = self.clicked(id, rect)
        canvas = self.canvas
        event = self.event
        canvas.push()
        self.skin.draw_menu(state, canvas, rect, label, (hover, hot))
        rects = []
        selected = None
        was_hot = state.get("hot", False)
        if hot or was_hot:
            for i,o in enumerate(options):
                canvas.translate((0,rect[3]+5))
                if hot:
                    self.skin.draw_menu_item(state, canvas, rect, (i, o))
                if was_hot and event.type == pygame.MOUSEBUTTONUP:
                    hit = self.canvas.rect(rect).collidepoint(event.pos)
                    if hit:
                        selected = i
        canvas.pop()
        state.hot = hot
        return selected

    @classmethod
    def input(self, name, rect):
        id = self.ID(name)
        state = self.STATE(name)
        text = state.get("text", [])
        cursor = state.get("cursor", 0)
        hover, hot, active = self.clicked(id, rect)
        canvas = self.canvas
        event = self.event
        pick = None
        if (hot or active):
            if event.type == pygame.MOUSEBUTTONDOWN:
                pick = canvas.tx(event.pos)
            if event.type == TICK:
                repeat = state.get("repeat")
                if repeat is not None:
                    revent, T = repeat
                    if time.time() - T > 0.05:
                        state.repeat = revent, time.time()
                        pygame.event.post(revent)    
            if active and event.type == pygame.KEYUP:
                state.repeat = None
        else:
            state.repeat = None
        if active and event.type == pygame.KEYDOWN:
            if state.get("repeat", None):
                state.repeat = event, time.time()
            else:
                state.repeat = event, time.time() + 0.5
            if event.key > 31 and event.key < 255:
                text.insert(cursor, event.unicode)
                cursor += 1
            else:
                if event.key == pygame.K_LEFT:
                    cursor -= 1
                if event.key == pygame.K_RIGHT:
                    cursor += 1
                if event.key == pygame.K_HOME:
                    cursor = 0
                if event.key == pygame.K_END:
                    cursor = len(text)
                if event.key == pygame.K_BACKSPACE:
                    cursor -= 1
                    if cursor >= 0:
                        text.pop(cursor)
                if event.key == pygame.K_RETURN:
                    GUI.active = False
                if cursor < 0: cursor = 0 
                if cursor > len(text): cursor = len(text)

        state.text = text
        state.cursor = cursor

        canvas.push()
        t = "".join(text)
        self.skin.draw_input(state, canvas, rect, (hover, hot, active), t, cursor, pick)
        canvas.pop()
        return t
        

